@@ -100,6 +100,11 @@ NLMSG_OVERRUN = 0x4    # Data lost
 NLMSG_MIN_TYPE = 0x10    # < 0x10: reserved control messages
 NLMSG_MAX_LEN = 0xffff  # Max message length
 
+mtypes = {1: 'NLMSG_NOOP',
+          2: 'NLMSG_ERROR',
+          3: 'NLMSG_DONE',
+          4: 'NLMSG_OVERRUN'}
+
 IPRCMD_NOOP = 0
 IPRCMD_STOP = 1
 IPRCMD_ACK = 2
@@ -207,6 +212,7 @@ class Marshal(object):
     '''
 
     msg_map = {}
+    debug = False
 
     def __init__(self):
         self.lock = threading.Lock()
@@ -245,7 +251,7 @@ class Marshal(object):
 
                 self.buf.seek(offset)
                 msg_class = self.msg_map.get(msg_type, nlmsg)
-                msg = msg_class(self.buf)
+                msg = msg_class(self.buf, debug=self.debug)
                 try:
                     msg.decode()
                     msg['header']['error'] = error
@@ -256,6 +262,9 @@ class Marshal(object):
                     msg['header']['error'] = e
                 except NetlinkDecodeError as e:
                     msg['header']['error'] = e
+                mtype = msg['header'].get('type', None)
+                if mtype in (1, 2, 3, 4):
+                    msg['event'] = mtypes.get(mtype, 'none')
                 self.fix_message(msg)
                 offset += msg.length
                 result.append(msg)
@@ -295,6 +304,10 @@ class PipeSocket(object):
     def fileno(self):
         return self.rfd
 
+    def close(self):
+        os.close(self.rfd)
+        os.close(self.wfd)
+
 
 def pairPipeSockets():
     pipe0_r, pipe0_w = os.pipe()
@@ -344,7 +357,6 @@ class IOThread(threading.Thread):
         self.pid = os.getpid()
         self._nonce = 0
         self._nonce_lock = threading.Lock()
-        self._stop = False
         self._run_event = threading.Event()
         self._sctl_event = threading.Event()
         self._stop_event = threading.Event()
@@ -406,15 +418,17 @@ class IOThread(threading.Thread):
         else:
             logging.error("got err for sctl, shutting down")
             # FIXME: shutdown all
-            self._stop = True
+            self._stop_event.set()
 
     def _feed_buffers(self):
         '''
         Beckground thread to feed reassembled buffers to the parser
         '''
         save = None
-        while not self._stop:
+        while True:
             (buf, marshal, sock) = self.buffers.get()
+            if self._stop_event.is_set():
+                return
             if save is not None:
                 # concatenate buffers
                 buf.seek(0)
@@ -446,13 +460,15 @@ class IOThread(threading.Thread):
         '''
         Background thread that expires masquerade cache entries
         '''
-        while not self._stop:
+        while True:
             # expire masquerade records
             ts = time.time()
             for i in tuple(self.masquerade.keys()):
                 if (ts - self.masquerade[i].ctime) > 60:
                     del self.masquerade[i]
-            time.sleep(60)
+            self._stop_event.wait(60)
+            if self._stop_event.is_set():
+                return
 
     def nonce(self):
         with self._nonce_lock:
@@ -516,10 +532,20 @@ class IOThread(threading.Thread):
             rsp['cmd'] = IPRCMD_ERR
             cmd = self.parse_control(data)
             if cmd['cmd'] == IPRCMD_STOP:
-                # Stop iothread
-                self._stop = True
-                self._stop_event.set()
+                # Last 'hello'
                 rsp['cmd'] = IPRCMD_ACK
+                rsp.encode()
+                sock.send(rsp.buf.getvalue())
+                # Stop iothread -- shutdown sequence
+                self._stop_event.set()
+                self._rlist.remove(self.sctl)
+                self._wlist.remove(self.sctl)
+                self.sctl.close()
+                self.control.close()
+                self.buffers.put((None, None, None))
+                self._feed_thread.join()
+                self._expire_thread.join()
+                return
             elif cmd['cmd'] == IPRCMD_RELOAD:
                 # Reload io cycle
                 self._reload_event.set()
@@ -605,7 +631,13 @@ class IOThread(threading.Thread):
             if self.record:
                 self.backlog.append((time.asctime(), msg))
             key = msg['header']['sequence_number']
-            msg['header']['host'] = _repr_sockets([sock], 'remote')[0]
+            try:
+                msg['header']['host'] = _repr_sockets([sock], 'remote')[0]
+            except socket.error:
+                # on shutdown, we can get here socket.error
+                # in this case add just an empty string
+                msg['header']['host'] = ''
+
             if key not in self.listeners:
                 key = 0
             if self.mirror and (key != 0) and (msg.raw is not None):
@@ -636,7 +668,10 @@ class IOThread(threading.Thread):
         return rsp
 
     def stop(self):
-        return self.command(IPRCMD_STOP)
+        try:
+            self.command(IPRCMD_STOP)
+        except OSError:
+            pass
 
     def reload(self):
         '''
@@ -715,13 +750,14 @@ class IOThread(threading.Thread):
     def start(self):
         threading.Thread.start(self)
         self._sctl_event.wait(3)
+        self._sctl_thread.join()
         if not self._sctl_event.is_set():
-            self._stop = True
-            raise Exception('failed to establish control connection')
+            self._stop_event.set()
+            raise RuntimeError('failed to establish control connection')
 
     def run(self):
         self._run_event.set()
-        while not self._stop:
+        while not self._stop_event.is_set():
             try:
                 [rlist, wlist, xlist] = select.select(self._rlist, [], [])
             except:
@@ -815,6 +851,7 @@ class Netlink(object):
         self.iothread.families[self.family] = self.send
         self.iothread.start()
         self.debug = debug
+        self.marshal.debug = debug
         if do_connect:
             self.connect(host, key, cert, ca)
 
@@ -911,13 +948,6 @@ class Netlink(object):
                 except:
                     pass
 
-    def shutdown(self):
-        '''
-        Deprecated: use Netlink.release() instead
-        '''
-        logging.warn('IPDB: using deprecated call "shutdown()"')
-        self.release()
-
     def release(self):
         '''
         Shutdown all threads and release netlink sockets
@@ -925,7 +955,8 @@ class Netlink(object):
         self.shutdown_sockets()
         self.shutdown_clients()
         self.shutdown_servers()
-        self._remote_cmd(self.iothread.control, IPRCMD_STOP)
+        self.iothread.stop()
+        self.iothread.join()
 
     def serve(self, url, key=None, cert=None, ca=None):
         if key:
